﻿-- maximum number of players that can be on the list of those synced through whispers
local SYNC_MAX_FRIENDS = 10

BMT = LibStub("AceAddon-3.0"):NewAddon("Black Market Tracker", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0")
BMTConfig = nil
BMTGlobal = nil

local defaults = {
	profile = {
		-- selected realm for multiple realm support
		selectedrealm = nil,
		-- controls sending and receiving messages in guild
		syncInGuild = true,
		-- controls sending and receiving messages in using friends
		syncWithFriends = true,
		-- friends synchronized with (in/out)
		syncedFriends = {},
	},
	global = {
		-- current items database
		current = {
			-- realm name
			['*'] = {
				updateInfo = {
					playerName = "Un_known",
					playerGuild = "",
					playerClass = "",
					updateDate = "01/01/12",
					updateTime = 0,
				},
				items = {
					-- item 1/2/3...
					['*'] = {
						itemID = 0,
						timeLeft = 0,
						predictedLow = 0,
						predictedHigh = 0,
						-- minutes after search (used for predictions calculations)
						predictionsUpdateDiff = 0,
						marketID = 0,
						-- date and time of last timeLeft change (used for removal criteria)
						lastTimeLeftChangeDate = "01/01/12",
						lastTimeLeftChangeDate = 0,
						minBid = 0,
						currBid = 0,
						minIncrement = 0,
						numBids = 0,
						quantity = 1,
						highestBidder = "",
					},
				},
			}
		},
		--[[
		-- item templates (items added here aren't stored in saved vars, new items are)
		templates = {
			-- itemID
			['*'] = {
				-- name of the item
				name = "Unknown",
				-- item type column
				itemType = "Unknown",
				-- "Level" column in bmah
				level = 1,
				-- item link
				-- generated automatically
				-- icon (texture path generated automatically)
				icon = "inv_misc_questionmark",
				-- quality (3 - rare, 4 - epic), needed for proper link coloring
				quality = 1,
			},
			[93194] = {
				["name"] = "Blood-Soaked Invitation",
				["itemType"] = "Quest",
				["level"] = 90,
				["icon"] = "inv_misc_bandage_05",
				["quality"] = 3,
			},
		},
		]]
	},
}

local options = {
	type = "group",
	childGroups = "tab",
	args = {
		info = {
			order = 1,
			name = "Addon Information",
			desc = "Shows informations about the addon",
			type = "execute",
			func = function()
				BMT:Print("This addon keeps track of Black Market Auction House auctions. Made by Jackk @ Auchindoun - EU")
			end,
		},
		main = {
			type = "group",
			name = "Main Module",
			args = {
				mainwindow = {
					order = 10,
					type = "header",
					name = "Main Window\n",
					cmdHidden = true
				},
				warnings = {
					order = 20,
					type = "header",
					name = "Warnings\n",
					cmdHidden = true
				},
				notifications = {
					order = 30,
					type = "header",
					name = "Notifications\n",
					cmdHidden = true
				},
			},
		},
		synchronization = {
			type = "group",
			name = "Synchronization",
			args = {
				channels = {
					order = 10,
					type = "header",
					name = "Channels\n",
					cmdHidden = true
				},
				syncguild = {
					order = 11,
					name = "Synchronize in Guild",
					desc = "Allow the addon to send and receive messages from players in your Guild.",
					type = 'toggle',
					get = function(info) return BMTConfig.syncInGuild end,
					set = function(info, x) 
					BMTConfig.syncInGuild = x
					end,
				},
				syncfriends = {
					order = 11,
					name = "Synchronize with Friends",
					desc = "Allow the addon to send and receive messages from pre-defined Friends.",
					type = 'toggle',
					get = function(info) return BMTConfig.syncWithFriends end,
					set = function(info, x)
					BMTConfig.syncWithFriends = x
					if x == false then
						BMTSideInset.TopText:SetText("Friends Synchronization (disabled)")
					else
						BMTSideInset.TopText:SetText("Friends Synchronization")
					end
					end,
				},
				friends = {
					order = 20,
					type = "header",
					name = "Friends\n",
					cmdHidden = true
				},
				friendsnote = {
					order = 21,
					type = "description",
					name = "NOTE: All friends that you want to synchronize your auctions with must be on your friends list. There can be up to "..SYNC_MAX_FRIENDS.." players on the list.\n",
				},
				manage = {
					order = 22,
					name = "Manage",
					desc = "Manage friends that you want to synchronize your auctions with.",
					type = "execute",
					func = function()
						InterfaceOptionsFrame:Hide()
						BMTFrame:Show()
						BMTFrame_Expand()
					end,
				},
			},
		},
	},
}

-- false before logging in, true after receiving friends info from the server
BMTGotFriendsInfo = false
-- function schedluer frame
local BMTSchedulerFrame = CreateFrame("Frame");
-- time since last call of the OnUpdate
local BMTElapsed = 0;
BMTScheduledTasks = {};
-- OnUpdate calls throttle
local BMT_ONUPDATE_SEC = 0.5;
-- set to true for additional dev messages
local BMTDebug = false;
--[===[@debug@
BMTDebug = true;
--@end-debug@]===]
-- this revision number
--BMTRevision = tonumber(GetAddOnMetadata("BlackMarketTracker", "Version"));
BMTRevision = "60"
if string.len(BMTRevision) == string.len("60") then
	BMTRevision = 0;
else
	BMTRevision = tonumber(BMTRevision)
end
-- addon name
local BMT_ADDON_NAME = "Black Market Tracker";
-- addon name without spaces
local BMT_ADDON_NAME_NO_SPACES = "BlackMarketTracker";
-- title text on top of the browser window
local BMT_TITLE = "Black Market Tracker";
-- new addon version text shown at the bottom of the browser
local BMT_NEW_VERSION = "New addon version available";
-- coming soon text
local BMT_COMING_SOON = "Coming Soon!";
-- more than 1 week ago text (last update)
local BMT_MORE_THAN_WEEK = "More than 1 week";
-- less than 1 minute ago text (last update)
local BMT_LESS_THAN_MINUTE = "Less than 1 minute";
-- completed auction text
local BMT_COMPLETED = "Completed!";
-- alpa for fading bottom text, goes above 1.0 so that the text stays lit for a while
local BMT_TEXT_ALPHA = 5.0;
-- time of fade in/out
local BMT_TEXT_TIME = 2;
-- predicted time left constants (time in minutes) - completed, short, medium, long, very long
local BMT_TIME_LEFT = {0, 30, 120, 720, 1440};
-- times left
local TIME_LEFT_COMPLETED = 0
local TIME_LEFT_LESS_THAN_30M = 1
local TIME_LEFT_30M_2H = 2
local TIME_LEFT_2H_12H = 3
local TIME_LEFT_MORE_THAN_12H = 4
-- array storing current db fields that are synced
local BMTItemSyncFields = 	{ "name",	"texture",	"quantity",	"itemType",	"level",	"levelType",	"sellerName",	"minBid",	"minIncrement",	"currBid",	"numBids",	"timeLeft",	"link",	"predictedLow",	"predictedHigh"	};
local BMTIsSyncFieldNumber =	{ false,	false,		true,		false,		true,		false,			false, 			true,		true,			true,		true,		true,		false,	true,			true			};
-- fading state (bottom text)
BMTFadeState = 0;

-- for expandable sideframe
local frameExpanded = false;
local selectedSideTab = "current"
local FRAME_WIDTH_COLLAPSED = 900;
local FRAME_WIDTH_EXPANDED = 1170;

-- popup text messages
-- forced wipe
local POPUP_OUTDATED_DATA_STRUCT_TEXT = "Data structure of "..BMT_ADDON_NAME.." is outdated and needs to be re-initialized.\n\nNOTE: This will reload your UI and wipe your current BMT data."
-- recommended current db wipe
local POPUP_RECOMMENDED_CURRENT_WIPE_TEXT = "It is recommended to wipe current items database of "..BMT_ADDON_NAME..". Not doing so may result in UI errors. \nIf you would like to do it later, please type '/bmt delete current'.\n\nNOTE: This will reload your UI."
local POPUP_CONFIRM_DELETE_ALL = "Are you sure you want to delete all of your data and settings?\n\nNOTE: Pressing 'Yes' will also reload your UI."
local POPUP_CONFIRM_DELETE_CURRENT = "Are you sure you want to delete current items database?\n\nNOTE: Pressing 'Yes' will also reload your UI."


function BMT:OnInitialize()
	defaults.profile.selectedrealm = GetRealmName()
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Black Market Tracker", "Black Market Tracker")
	LibStub("AceConfig-3.0"):RegisterOptionsTable(BMT_ADDON_NAME, options)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(BMT_ADDON_NAME, options)
	self.db = LibStub("AceDB-3.0"):New("BlackMarketTrackerDB", defaults, true)
	BMTConfig = self.db.profile
	BMTGlobal = self.db.global
	options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
end

function BMT:OnEnable()
	--self:Print("Black Market Tracker Loaded!")
	--self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	
	-- in syncing module, has to be done here, as otherwise this function cannot be seen
	BMTSpinningInset_Initialize()
	-- query server for friends list
	-- TODO: BNET
	ShowFriends()
	-- version is synced after receiving response from the server
end

function BMT:OnDisable()
end

function BMT:RefreshConfig()
	BMTConfig = self.db.profile
	global = self.db.global
end


function BMT:Debug(msg)
	if(BMTDebug) then
		local text = "|cff00ff96(Debug): |r"..msg;
		BMT:Print(text);
	end
end

function BMT_SlashDebugHandler()
	
end

BMTSchedulerFrame:SetScript("OnUpdate", function(self, elapsedVal)
	-- throttle BMT_OnUpdate calls
	BMTElapsed = BMTElapsed + elapsedVal;
	if(BMTElapsed >= BMT_ONUPDATE_SEC) then
		BMTElapsed = 0;
		BMT_OnUpdate(self, elapsedVal);
	end
end)

function BMT_DeleteAllData()
	BMTGlobal.current = nil
	ReloadUI();
end

function BMT_DeleteCurrentDB()
	BMTGlobal.current = nil
	ReloadUI()
end

function BMT_SlashHandler(msg, editbox)
	local msgLen = #msg;
	-- if any popup window is visible, ignore everything
	if BMTIsAnyPopupWindowShown() then return end
	
	if(msgLen == 0 or msg == "show" or msg == "ui" or msg == "open") then
		BMT:ToggleFrame();
	elseif(msg == "debug") then
		BMT_SlashDebugHandler();
	elseif(msg == "delete") then
		BMTShowPopupWindowConfirm(BMT_CONFIRM_WINDOW_STATE_DELETE_ALL, POPUP_CONFIRM_DELETE_ALL)
	elseif(msg == "delete current") then
		BMTShowPopupWindowConfirm(BMT_CONFIRM_WINDOW_STATE_DELETE_CURRENT, POPUP_CONFIRM_DELETE_CURRENT)
	elseif(msg == "config") then
		InterfaceOptionsFrame_OpenToCategory(BMT_ADDON_NAME)
	elseif(msg == "sync" or msg == "synchronize") then
		BMTSyncing_BroadcastAuctionsRequest()
	end
end

function BMT_EnableEscInteractivity(frameName)
	local found = false
	for i=1, #UISpecialFrames do
		if UISpecialFrames[i] == frameName then
			found = true
			break
		end
	end
	if not found then 
		table.insert(UISpecialFrames, frameName)
	end
end

function BMT_DisableEscInteractivity(frameName)	
	for i=1, #UISpecialFrames do	
		if UISpecialFrames[i] == frameName then
			table.remove(UISpecialFrames, i)
			break
		end
	end
end

function BMTFrame_OnLoad(self)
	SLASH_BLACKMARKETTRACKER1 = "/bmt";
	SLASH_BLACKMARKETTRACKER2 = "/blackmarkettracker";
	SlashCmdList["BLACKMARKETTRACKER"] = BMT_SlashHandler;
	self:RegisterForDrag("LeftButton");
	BMT_EnableEscInteractivity(self:GetName())
	self:RegisterEvent("BLACK_MARKET_ITEM_UPDATE");
	self:RegisterEvent("FRIENDLIST_UPDATE");
	self:RegisterEvent("GET_ITEM_INFO_RECEIVED");
	BMTScrollFrame.update = BMTScrollFrame_Update;
	BMTScrollFrame.scrollBar.doNotHide = true;
	BMTFrameTitleText:SetText(BMT_TITLE.." (r"..BMTRevision..")");
	RegisterAddonMessagePrefix("BMT");
end

function BMTFrame_OnEvent(self, event, ...)
	-- item update throttling/fix
	if(event == "BLACK_MARKET_ITEM_UPDATE") then
		BMTTaskManager_Unschedule(BMT_NewSearchHandler);
		BMTTaskManager_Schedule(0.5, BMT_NewSearchHandler);
	elseif(event == "GET_ITEM_INFO_RECEIVED") then
		BMTTaskManager_Unschedule(BMT_FrameRefresh);
		BMTTaskManager_Schedule(0.5, BMT_FrameRefresh);
	elseif(event == "FRIENDLIST_UPDATE") then
		BMT_FriendlistUpdateHandler();
	end
end

function BMT_FriendlistUpdateHandler()
	if not BMTGotFriendsInfo then
		-- sync version
		-- UNCOMMENT
		--BMTSyncing_AskForVersion_OnLogin()
		BMTGotFriendsInfo = true
	end
	if BMTSideInset:IsShown() then
		BMTSideInsetScrollFrame_Update()
	end
	-- syncing callback
	if BMTBroadcastQueued then
		-- UNCOMMENT
		--BMTSyncing_PerformRealBroadcast()
	end
end

function BMT:ToggleFrame()
	if(BMTFrame:IsShown()) then
		BMTFrame:Hide();
		--BlackMarketTrackerBottom:SetAlpha(0);
		-- TODO: sound enable toggle
	else
		BMTFrame:Show();
		-- TODO: sound enable toggle
	end
end

-- under development overlay and buttons disable
function BMTFrame_MakeUnusable()
	BMTFrame.UnderDevelopment:Show()
	BMTFrame.UnderDevelopment:EnableMouse(true)
	BMTFrame:EnableMouse(false);
	local buttons = BMTScrollFrame.buttons;
	local numButtons = 0;
	if buttons ~= nil then
		numButtons = #buttons;
	end
	for i=1, numButtons do
		buttons[i]:EnableMouse(false);
		buttons[i].Item:EnableMouse(false);
		buttons[i].TimeLeft:EnableMouse(false);
	end
end

function BMT_FrameRefresh()
	-- add tabs support later
	if BMTFrame:IsShown() then
		-- if selectedTab == ... then
		BMTScrollFrame_Update()
	end
end

function BMTFrame_OnShow()
	if(BMTHighestRevision <= BMTRevision) then
		BMTFadeState = 2;
	else
		BMTFadeState = 0;
	end
	BMTBottomText_FadingHandler()
	PlaySound("AuctionWindowOpen")
	HybridScrollFrame_CreateButtons(BMTScrollFrame, "BMTItemTemplate", 5, -5)
	-- under development overlay
	if not BMTDebug then
		BMTFrame_MakeUnusable()
	end
	BMT_RefreshAllPredictedTimes(BMTConfig.selectedrealm)
	BMTScrollFrame_Update()
	-- be careful about changing this number
	if BMTSyncingState == 0 then
		BMTFrameSpinningInset:Hide()
	else
		BMTFrameSpinningInset:Show()
	end
end

-- expandable side frame stuff
function BMTSideInset_OnLoad(self)
	self.ScrollFrame.update = BMTSideInsetScrollFrame_Update
	self.ScrollFrame.scrollBar.doNotHide = true
end

function BMTSideInset_OnShow(self)
	-- query server for friends
	-- TODO: BNET
	ShowFriends()
	-- scroll frame
	HybridScrollFrame_CreateButtons(self.ScrollFrame, "BMTFriendsRowTemplate", -30, -5)
	BMTSideInsetScrollFrame_Update()
	-- set current 'tab' to be the chosen one
	selectedSideTab = "current"
	BMTSideInsetButtonCurrent:GetNormalTexture():SetAlpha(0.5);
	BMTSideInsetButtonAddNew:GetNormalTexture():SetAlpha(0);
	-- update bottom text
	
	-- top text
	if BMTConfig.syncWithFriends then
		BMTSideInset.TopText:SetText("Friends Synchronization")
	else
		BMTSideInset.TopText:SetText("Friends Synchronization (disabled)")
	end
	-- bottom text
	BMTSideInset_BottomTextUpdate()
end

function BMTFrame_IsExpanded()
	return frameExpanded
end

function BMTFrame_Collapse()
	frameExpanded = false
	BMTFrame:SetWidth(FRAME_WIDTH_COLLAPSED)
	BMTFrameExpandButton:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up")
	BMTFrameExpandButton:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down")
	BMTFrameExpandButton:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled")
	BMTSideInset:Hide();
end

function BMTFrame_Expand()
	frameExpanded = true;
	BMTFrame:SetWidth(FRAME_WIDTH_EXPANDED);
	BMTFrame.ExpandButton:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up");
	BMTFrame.ExpandButton:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down");
	BMTFrame.ExpandButton:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled");
	BMTSideInset:Show();
end

function BMTSideInsetCurrent_OnClick(self)
	BMTSideInset.selectedFriendName = nil
	self:GetNormalTexture():SetAlpha(0.5)
	selectedSideTab = "current"
	BMTSideInsetButtonAddNew:GetNormalTexture():SetAlpha(0)
	BMTSideInsetScrollFrame_Update()
	BMTSideInset.AddRemoveButton:SetText("Remove")
end

function BMTSideInsetAddNew_OnClick(self)
	BMTSideInset.selectedFriendName = nil
	self:GetNormalTexture():SetAlpha(0.5)
	selectedSideTab = "addnew"
	BMTSideInsetButtonCurrent:GetNormalTexture():SetAlpha(0)
	BMTSideInsetScrollFrame_Update()
	BMTSideInset.AddRemoveButton:SetText("Add")
end

function BMT_IsOnFriendsList(name)
	-- TODO: BNET
	local online
	local numFriends = GetNumFriends()
	for i=1, numFriends do
		local n, _, _, _, online = GetFriendInfo(i)
		if n == name then return true end
	end
	return false
end

function BMT_IsFriendOnline(name)
	-- TODO: BNET
	local n, online
	local numFriends = GetNumFriends()
	for i=1, numFriends do
		n, _, _, _, online = GetFriendInfo(i)
		if n == name then return online end
	end
	return nil
end

function BMTSideInsetAddRemoveButton_OnClick(self)
	if selectedSideTab == "current" then
		-- removal
		local index = 0
		for i=1, #BMTConfig.syncedFriends do
			if BMTConfig.syncedFriends[i] == BMTSideInset.selectedFriendName then
				index = i
				break
			end
		end
		if index > 0 then
			table.remove(BMTConfig.syncedFriends, index)
		end
	elseif selectedSideTab == "addnew" then
		-- addition
		-- necessary safety checks for duplicates handled by disabling the button correctly
		if BMTSideInset.selectedFriendName ~= nil then
			table.insert(BMTConfig.syncedFriends, BMTSideInset.selectedFriendName)
		end
		table.sort(BMTConfig.syncedFriends)
	end
	BMTSideInsetScrollFrame_Update()
	BMTSideInset_BottomTextUpdate()
end

function BMTSideInset_BottomTextUpdate()
	BMTSideInset.BottomText:SetText("NOTE: You can synchronize with up to "..SYNC_MAX_FRIENDS.." friends at a time. Currently using: "..#BMTConfig.syncedFriends.." out of "..SYNC_MAX_FRIENDS..".\n\nIn order to synchronize, you must have your friends added to your 'current' list and vice versa (messages from strangers are ignored).\n\nOnly players added to your in-game friends list can be added to 'current' list and synchronized with.")
end

function BMTSideInsetScrollFrameButton_OnClick(self)
	BMTSideInset.selectedFriendName = self.Name:GetText()
	BMTSideInsetScrollFrame_Update()
end

function BMTSideInsetScrollFrame_Update()
	local numFriends
	if selectedSideTab == "current" then
		numFriends = #BMTConfig.syncedFriends
	elseif selectedSideTab == "addnew" then
		numFriends = GetNumFriends()
	end
	local scrollFrame = BMTSideInsetScrollFrame;
	local offset = HybridScrollFrame_GetOffset(scrollFrame);
	local buttons = scrollFrame.buttons;
	local numButtons = #buttons;
	local enableButton = false;
	for i=1, numButtons do
		local button = buttons[i];
		local index = offset + i;
		if index <= numFriends then
			local name
			if selectedSideTab == "current" then
				name = BMTConfig.syncedFriends[i]
			elseif selectedSideTab == "addnew" then
				name = GetFriendInfo(index)
			end
			if name then
				if selectedSideTab == "current" then
					if BMT_IsOnFriendsList(name) then
						-- online / offline
						if BMT_IsFriendOnline(name) then
							button.Name:SetText(name)
							button.Name:SetTextColor(1.0, 1.0, 1.0)
						else
							button.Name:SetText(name.. " (Offline)")
							button.Name:SetTextColor(0.6, 0.6, 0.6)
						end
					else
						-- not on friends list
						button.Name:SetText(name.. " (Not on friends list)")
						button.Name:SetTextColor(1.0, 0.1, 0.1)
					end
				elseif selectedSideTab == "addnew" then
					button.Name:SetText(name)
					button.Name:SetTextColor(1.0, 1.0, 1.0)
				end
				button:Show()
				if name == BMTSideInset.selectedFriendName then
					button.Selection:Show();
					-- for add new tab enable button only if selection isn't already added
					if selectedSideTab == "addnew" then
						local foundInCurrent = false
						for j=1, numFriends do
							if name == BMTConfig.syncedFriends[j] then
								foundInCurrent = true
								break
							end
						end
						if not foundInCurrent then
							enableButton = true
						end
					elseif selectedSideTab == "current" then
						enableButton = true
					end
				else
					button.Selection:Hide();
				end
			else
				button:Hide()
			end
		else
			button:Hide()
		end
	end
	-- button handling, button is active only when selection is active
	if enableButton and #BMTConfig.syncedFriends < SYNC_MAX_FRIENDS then
		BMTSideInset.AddRemoveButton:Enable()
	else
		BMTSideInset.AddRemoveButton:Disable()
	end
	local totalHeight = numFriends * scrollFrame.buttonHeight;
    local displayedHeight = numButtons * scrollFrame.buttonHeight;
    HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);
end

function BMTFrame_OnHide()
	PlaySound("AuctionWindowClose");
end

-- delay_time, function, function_parameters
function BMTTaskManager_Schedule(delay, func, ...)
	-- add params
	local task = {...};
	-- add function and time when it should be called
	task.func = func;
	task.time = GetTime() + delay;
	table.insert(BMTScheduledTasks, task);
end

function BMTTaskManager_Unschedule(func, ...)
	local entry, found;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.func == func) then
			found = true;
			for j=1, select("#", ...) do
				if(select(j, ...) ~= entry[i]) then
					found = false;
					break;
				end
			end
			if(found) then
				table.remove(BMTScheduledTasks, i);
			end
		end
	end
end

function BMT_OnUpdate(self, elapsedVal)
	local entry;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.time <= GetTime()) then
			table.remove(BMTScheduledTasks, i);
			entry.func(unpack(entry));
		end
	end
end

function BMTItem_OnClick(self, button, down)
	-- Function for later use
end

function BMTBottomText_FadingHandler()
	-- -1 is only used if the window is currently opened and new version was found during that
	-- therefore we want to fade out current text first and then fade into the new one
	if(BMTFadeState == -1) then
		BMTFadeState = 0;
		BMTBottomText_FadeOut();
	elseif(BMTFadeState == 0) then
		BMTFadeState = 1;
		BMTFrameBottomText:SetText(BMT_NEW_VERSION .. " (r" .. BMTHighestRevision .. ")");
		BMTTrackerBottomText_FadeIn();
	elseif(BMTFadeState == 1) then
		BMTFadeState = 2;
		BMTBottomText_FadeOut();
	elseif(BMTFadeState == 2) then
		BMTFadeState = 3;
		local bottomText = "Last updated by ";
		if BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.playerName == "Un_known" then
			bottomText = bottomText .. "|cffffd100Unknown|r";
		else
			local classColor = RAID_CLASS_COLORS[BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.playerClass];
			bottomText = bottomText .. "|cff" .. string.format("%2.2x", classColor.r*255) .. string.format("%2.2x", classColor.g*255) .. string.format("%2.2x", classColor.b*255);
			bottomText = bottomText .. BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.playerName .. "|r (";
			bottomText = bottomText .. BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.playerGuild .. "), ";
			local updateTimeStr =  string.format("%2.2d:%2.2d", (BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.updateTime/60), (BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.updateTime%60));
			bottomText = bottomText .. BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.updateDate .. " " .. updateTimeStr.." (";
			local currHour, currMinute = GetGameTime();
			local currTime = 60*currHour+currMinute;
			local diff = BMT_CalculateDateAndTimeDiff(BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.updateDate, BMTGlobal.current[BMTConfig.selectedrealm].updateInfo.updateTime, date("%d/%m/%y"), currTime);
			-- 1 week
			if(diff > 10080) then
				bottomText = bottomText .. BMT_MORE_THAN_WEEK;
			elseif(diff < 1) then
				bottomText = bottomText .. BMT_LESS_THAN_MINUTE;
			else
				local addSpace = false;
				local days = math.floor(diff/1440);
				local hours = math.floor((diff-(days*1440))/60);
				local minutes = math.floor(diff%60);
				if(days ~= 0) then
					addSpace = true;
					bottomText = bottomText .. tostring(days);
					if(days == 1) then
						bottomText = bottomText .. " day";
					else
						bottomText = bottomText .. " days";
					end						
				end
				
				if(hours ~= 0) then
					if(addSpace) then
						bottomText = bottomText .." ";
					end
					addSpace = true;
					bottomText = bottomText .. tostring(hours);
					if(hours == 1) then
						bottomText = bottomText .. " hour";
					else
						bottomText = bottomText .. " hours";
					end						
				end
				
				if(minutes ~= 0) then
					if(addSpace) then
						bottomText = bottomText .." ";
					end
					addSpace = true;
					bottomText = bottomText .. tostring(minutes);
					if(minutes == 1) then
						bottomText = bottomText .. " minute";
					else
						bottomText = bottomText .. " minutes";
					end						
				end
			end
			bottomText = bottomText .. " ago)";
		end
		BMTFrameBottomText:SetText(bottomText);
		BMTTrackerBottomText_FadeIn();
	end
end

function BMTTrackerBottomText_FadeIn()
	local fadeInfo = {};
    fadeInfo.mode = "IN";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.endAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BMTFrameBottom, fadeInfo);
end

function BMTBottomText_FadeOut()
	local fadeInfo = {};
    fadeInfo.mode = "OUT";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.startAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BMTFrameBottom, fadeInfo);
end

function BMTScrollFrame_Update()
	local numItems = BMT_GetNumItems(BMTConfig.selectedrealm)
	local scrollFrame = BMTScrollFrame;
	local offset = HybridScrollFrame_GetOffset(scrollFrame);
	local buttons = scrollFrame.buttons;
	local numButtons = #buttons;
	
	for i=1, numButtons do
		local button = buttons[i];
		local index = offset + i;
		if (index <= numItems) then
			local itemID = BMTGlobal.current[BMTConfig.selectedrealm].items[index].itemID
			local quantity = BMTGlobal.current[BMTConfig.selectedrealm].items[index].quantity
			local numBids = BMTGlobal.current[BMTConfig.selectedrealm].items[index].numBids
			local highestBidder = BMTGlobal.current[BMTConfig.selectedrealm].items[index].highestBidder
			if highestBidder == "" then
				highestBidder = "---"
			end
			local minIncrement = BMTGlobal.current[BMTConfig.selectedrealm].items[index].minIncrement
			local currBid = BMTGlobal.current[BMTConfig.selectedrealm].items[index].currBid
			local timeLeft = BMTGlobal.current[BMTConfig.selectedrealm].items[index].timeLeft
			local predictedLow = BMTGlobal.current[BMTConfig.selectedrealm].items[index].predictedLow
			local predictedHigh = BMTGlobal.current[BMTConfig.selectedrealm].items[index].predictedHigh
			local minBid = BMTGlobal.current[BMTConfig.selectedrealm].items[index].minBid
			local name, link, _, ilvl, _, _, _, _, _, texture = GetItemInfo(itemID)
			if(itemID ~= 0) then
				-- GetItemInfo returned an item
				if(name) then
					button.Name:SetText(name)
					button.itemLink = link
					button.ItemLevel:SetText(ilvl)
					button.Item.IconTexture:SetTexture(texture)
				-- GetItemInfo returned nil, scrollframe will get updated once GET_ITEM_INFO_RECEIVED fires
				else
					button.Name:SetText("Unknown#"..itemID)
					button.itemLink = nil
					button.ItemLevel:SetText(1)
					button.Item.IconTexture:SetTexture("Interface\\Icons\\Inv_Misc_Questionmark")
				end
				-- config?
				--[[
                if ( not usable ) then
                    button.Item.IconTexture:SetVertexColor(1.0, 0.1, 0.1)
                else
                    button.Item.IconTexture:SetVertexColor(1.0, 1.0, 1.0)
                end
				]]
                button.Item.Count:SetText(quantity)
                button.Item.Count:SetShown(quantity > 1)
                button.Type:SetText(BMT_COMING_SOON)
				button.NumBids:SetText(numBids)
                button.HighestBidder:SetText(highestBidder)
                
                local bidAmount = currBid
                local minNextBid = currBid + minIncrement
                if ( currBid == 0 ) then
                    bidAmount = minBid
                    minNextBid = minBid
                end
                MoneyFrame_Update(button.CurrentBid, bidAmount)
                button.minNextBid = minNextBid
                button.auctionComplete = (timeLeft == 0)
                button.TimeLeft.Text:SetText(_G["AUCTION_TIME_LEFT"..timeLeft])
                button.TimeLeft.tooltip = _G["AUCTION_TIME_LEFT"..timeLeft.."_DETAIL"]
				local predictedText = BMT_PredictedTimeGetText(predictedLow, predictedHigh)
				button.PredictedTimeLeft:SetText(predictedText)
				button.Notification:SetText(BMT_COMING_SOON)
                button:Show()
            else
                button:Hide()
			end
		else
			button:Hide()
		end
	end
	local totalHeight = numItems * scrollFrame.buttonHeight
    local displayedHeight = numButtons * scrollFrame.buttonHeight
    HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight)
end

function BMT_PredictedTimeGetText(predictedLow, predictedHigh)
	local text = "";
	local predictedLowH = math.floor(predictedLow/60);
	local predictedLowM = predictedLow%60;
	local predictedHighH = math.floor(predictedHigh/60);
	local predictedHighM = predictedHigh%60;
	-- more than x
	if(predictedHigh > (BMT_TIME_LEFT[#BMT_TIME_LEFT-1]+1)) then
		text = "> ";
		if(predictedLowH ~= 0) then
			text = text..tostring(predictedLowH).."h";
		end
		if(predictedLowM ~= 0) then
			if(predictedLowH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedLowM).."m";
		end
	-- less than x
	elseif(predictedLow <= 0 and predictedHigh > 0) then
		text = "< ";
		if(predictedHighH ~= 0) then
			text = text..tostring(predictedHighH).."h";
		end
		if(predictedHighM ~= 0) then
			if(predictedHighH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedHighM).."m";
		end
	-- completed
	elseif(predictedLow <= 0 and predictedHigh <= 0) then
		text = BMT_COMPLETED;
	-- between x and y
	else
		if(predictedLowH ~= 0) then
			text = text..tostring(predictedLowH).."h";
		end
		if(predictedLowM ~= 0) then
			if(predictedLowH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedLowM).."m";
		end
		text = text .. " - ";
		if(predictedHighH ~= 0) then
			text = text..tostring(predictedHighH).."h";
		end
		if(predictedHighM ~= 0) then
			if(predictedHighH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedHighM).."m";
		end
	end
	return text;
end

function BMT_SetNewUpdater(realmName, playerName, playerClass, playerGuild, updateTime, updateDate)
	-- checks if table exists should be done before already
	BMTGlobal.current[realmName].updateInfo.playerName = playerName
	BMTGlobal.current[realmName].updateInfo.playerClass = playerClass
	BMTGlobal.current[realmName].updateInfo.playerGuild = playerGuild
	BMTGlobal.current[realmName].updateInfo.updateTime = updateTime
	BMTGlobal.current[realmName].updateInfo.updateDate = updateDate
	-- if tracker is opened already, update the bottom text
	if(BMTFrame:IsShown()) then
		BMTFadeState = 1
		BMTBottomText_FadingHandler()
	end
end

-- recalculates predicted times left for item under [index] for [realmName] realm
-- this also sets the calculated predicted times, predictionsUpdateDate and predictionsUpdateTime for that item
-- DOES NOT set other fields like timeLeft to new value
-- --parameters:
-- --realmName - name of the realm of a given item
-- --index - index of the item in realm's item db
-- --newPredictedLow, newPredictedHigh, newTimeLeft, newUpdateDate, newUpdateTime - correspond to the new item
-- --dateOfSearch, timeOfSearch - current date and current time should be passed, unless we're performing a 'fake' recalculation (one that isn't bringing predictions to current times, only done as a step of calculating predictions upon sync, without a new search), then date and time of incoming search should be passed. NOTE: make sure that dateOfSearch and timeOfSearch are not greater than date and time of our current search
-- --fromWindowRefresh - true if this function has been called by just refreshing the browser window, false otherwise (if false, everything going under or above blizzard's time range will be cut off)
-- this can be called either to recalculate the times upon browsing BMAH ourself, receiving a sync or just refreshing browser list (in which case the same timeLeft should be passed and fromWindowRefresh set to true)
-- upon receiving a sync 'negotiating' sync (finding narrowest possible prediction) this function will recalculate predictions, as if receiver was the one searching the items
function BMT_RecalculateAndSetPredictedTimes(realmName, index, newTimeLeft, dateOfSearch, timeOfSearch, fromWindowRefresh)
	-- correct errors, just in case they happen..
	if BMTGlobal.current[realmName].items[index].predictedLow > BMTGlobal.current[realmName].items[index].predictedHigh then
		BMT:Debug("prevPredictedLow > prevPredictedHigh for i="..index.." name="..BMTGlobal.current[realmName].items[index].itemID..". Was: predictedLow="..BMTGlobal.current[realmName].items[index].predictedLow.." predictedHigh="..BMTGlobal.current[realmName].items[index].predictedHigh..". Setting default predictedLow to "..BMT_TIME_LEFT[newTimeLeft])
		BMTGlobal.current[realmName].items[index].predictedLow = BMT_TIME_LEFT[newTimeLeft]
	end
	
	local prevPredictedLow = BMTGlobal.current[realmName].items[index].predictedLow
	local prevPredictedHigh = BMTGlobal.current[realmName].items[index].predictedHigh
	local prevTimeLeft = BMTGlobal.current[realmName].items[index].timeLeft
	local predPredUpdateMinAfterSearch = BMTGlobal.current[realmName].items[index].predictionsUpdateDiff
	local searchDate = BMTGlobal.current[realmName].updateInfo.updateDate
	local searchTime = BMTGlobal.current[realmName].updateInfo.updateTime
	local updatedMinAgo = BMT_CalculateDateAndTimeDiff(dateOfSearch, timeOfSearch, searchDate, searchTime) - predPredUpdateMinAfterSearch
	-- set those fields here already, as this should be done for all cases
	BMTGlobal.current[realmName].items[index].predictionsUpdateDiff = BMT_CalculateDateAndTimeDiff(dateOfSearch, timeOfSearch, searchDate, searchTime)
	-- completed case
	if newTimeLeft == TIME_LEFT_COMPLETED then
		BMTGlobal.current[realmName].items[index].predictedLow = BMT_TIME_LEFT[1] -- 0
		BMTGlobal.current[realmName].items[index].predictedHigh = BMT_TIME_LEFT[1] -- 0
		return
	end
	-- more than 12h case
	if newTimeLeft == TIME_LEFT_MORE_THAN_12H then return end
	-- more than completed, less than 12h
	-- if timeLeft changed, example:
	-- old: 2h-12h
	-- new: 30m-2h
	-- ..............30m.....................2h............................................12h	scale
	-- ......................................|<--------------old prediction------------>|....	old prediction
	-- .........................|<--------------old prediction------------>|.................	old prediction moved left by (updatedMinAgo=newUpdateTime-oldUpdateTime)
	-- .........................|<----new--->|...............................................	everything going under lower time range (30m) or above higher time range (2h) is cut off
	
	if newTimeLeft < prevTimeLeft then
		BMT:Debug("index="..index.." itemID="..BMTGlobal.current[realmName].items[index].itemID.." UP->DOWN. prevPredictedLow="..prevPredictedLow.." prevPredictedHigh="..prevPredictedHigh.." updatedMinAgo="..updatedMinAgo)
		-- never touch the upper range of 'more than 12h'
		if prevTimeLeft ~= TIME_LEFT_MORE_THAN_12H then
			prevPredictedHigh = prevPredictedHigh - updatedMinAgo
		end
		-- if timeLeft is >12h, lower prediction isn't ever moved, therefore we need to handle that accordingly to updateInfo date and time, without taking predictionsUpdateDiff into account
		if prevTimeLeft == TIME_LEFT_MORE_THAN_12H then
			prevPredictedLow = prevPredictedLow - BMT_CalculateDateAndTimeDiff(dateOfSearch, timeOfSearch, searchDate, searchTime)
		else
			-- lower range can go to negative, it will get cut / re-adjusted anyway
			prevPredictedLow = prevPredictedLow - updatedMinAgo
		end
		-- cut everything that goes under or above given (blizzard's) time left range
		if prevPredictedLow < BMT_TIME_LEFT[newTimeLeft] then
			prevPredictedLow = BMT_TIME_LEFT[newTimeLeft]
		end
		if prevPredictedHigh > BMT_TIME_LEFT[newTimeLeft+1]+1 then
			prevPredictedHigh = BMT_TIME_LEFT[newTimeLeft+1]+1
		end
		-- DEBUG
		if prevPredictedLow > prevPredictedHigh then
			BMT:Debug("prevPredictedLow > prevPredictedHigh (newTimeLeft < prevTimeLeft), prevPredictedLow="..prevPredictedLow.." prevPredictedHigh="..prevPredictedHigh)
		end
		-- END OF DEBUG
		-- if completed, then set both to 0
		if prevPredictedHigh <= 0 then
			prevPredictedLow = 0
			prevPredictedHigh = 0
		end
		-- set the fields
		BMTGlobal.current[realmName].items[index].predictedLow = prevPredictedLow
		BMTGlobal.current[realmName].items[index].predictedHigh = prevPredictedHigh
		return
	end
	
	-- if timeLeft didn't change or we're just refreshing the window, then we just cut (updatedMinAgo=newUpdateTime-oldUpdateTime) from the upper range
	if newTimeLeft == prevTimeLeft then
		-- never touch the upper range of 'more than 12h'
		if prevTimeLeft ~= TIME_LEFT_MORE_THAN_12H then
			prevPredictedHigh = prevPredictedHigh - updatedMinAgo
		end
		-- if we're just opening the BMT window, then we lower both, if searching we cut the lower range to lower timeLeft range
		prevPredictedLow = prevPredictedLow - updatedMinAgo
		if not fromWindowRefresh then
			if prevPredictedLow < BMT_TIME_LEFT[newTimeLeft] then
				prevPredictedLow = BMT_TIME_LEFT[newTimeLeft]
			end
		end
		-- DEBUG
		if prevPredictedLow > prevPredictedHigh then
			BMT:Debug("prevPredictedLow > prevPredictedHigh (newTimeLeft == prevTimeLeft), prevPredictedLow="..prevPredictedLow.." prevPredictedHigh="..prevPredictedHigh)
		end
		-- END OF DEBUG
		BMTGlobal.current[realmName].items[index].predictedLow = prevPredictedLow
		BMTGlobal.current[realmName].items[index].predictedHigh = prevPredictedHigh
		-- prevPredictedLow stays on the lower range of current given (blizzard's) time left
		return
	end
end

-- called only when refreshing browser, otherwise BMT_RecalculateAndSetPredictedTimes should be called directly for each item
function BMT_RefreshAllPredictedTimes(realmName)
	local numItems = BMT_GetNumItems(BMTConfig.selectedrealm)
	local currDate = date("%d/%m/%y");
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	for i=1, numItems do
		local currTimeLeft = BMTGlobal.current[realmName].items[i].timeLeft
		BMT_RecalculateAndSetPredictedTimes(realmName, i, currTimeLeft, currDate, currTime, true)
	end
end

-- moves predicted times of all items for given name back to the point of where they were when last search was done
-- this does not touch auctions with only lower range (>12h) or completed ones
function BMT_MoveAllPredictionsToLastSearchTime(realmName)
	local numItems = BMT_GetNumItems(BMTConfig.selectedrealm)
	local currDate = date("%d/%m/%y");
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	local searchDate = BMTGlobal.current[realmName].updateInfo.updateDate
	local searchTime = BMTGlobal.current[realmName].updateInfo.updateTime
	for i=1, numItems do
		local timeLeft = BMTGlobal.current[realmName].items[i].timeLeft
		if timeLeft ~= TIME_LEFT_COMPLETED and timeLeft ~= TIME_LEFT_MORE_THAN_12H then
			local diff = BMTGlobal.current[realmName].items[i].predictionsUpdateDiff
			BMTGlobal.current[realmName].items[i].predictionsUpdateDiff = 0
			BMTGlobal.current[realmName].items[i].predictedLow = BMTGlobal.current[realmName].items[i].predictedLow + diff
			BMTGlobal.current[realmName].items[i].predictedHigh = BMTGlobal.current[realmName].items[i].predictedHigh + diff
			-- no need to cut anything that goes above blizzard's given time range, as it has already been done upon searching
		end
	end
end

-- moves all outdated entries to history DB and removes them from current DB under given criteria:
-- a - date and time of last timeLeft change is greater than timeLeftHighRange+2h (for timeLeft >12h assuming that auction can't ever be up for longer than BMT_TIME_LEFT[highest] - current longest is 24h+2h)
-- b - auction with given marketID cannot be found among auctions from searching/syncing
-- c - auction with given marketID can be found among auctions from searching/syncing but has a different name or its time left went up (assuming auctions' time left don't get extended upon bidding) or gold went down or number of bids went down
----- parameters are:	- realmName to remove the items for
--						- foundItems - list of all items found by searching BMAH or from syncing
--						- dateOfSearch - date when the search was performed. If called when searching BMAH yourself, current date should be passed
--						- timeOfSearch - time when the search was performed. If called when searching BMAH yourself, current time should be passed
-- performs removal from current table
-- initializes current and history tables for a given realm if needed
function BMT_MoveAllOldEntriesToHistory(realmName, foundItems, dateOfSearch, timeOfSearch)
	-- criterion a
	local currDBSize = #BMTGlobal.current[realmName].items
	local i=1
	while i<=currDBSize do
		if BMT_CalculateDateAndTimeDiff(BMTGlobal.current[realmName].items[i].lastTimeLeftChangeDate, BMTGlobal.current[realmName].items[i].lastTimeLeftChangeTime, dateOfSearch, timeOfSearch) > BMT_TIME_LEFT[BMTGlobal.current[realmName].items[i].timeLeft+1]+120 then
			BMT_MoveEntryToHistory(realmName, i)
			currDBSize=currDBSize-1
			i=i-1
		end
		i=i+1
	end
	-- criteria b and c
	currDBSize = #BMTGlobal.current[realmName].items
	i=1
	while i<=currDBSize do
		local found = false
		local itemIDMatches = true
		local timeLeftWentUp = false
		local goldWentDown = false
		local numBidsWentDown = false
		for j=1, #foundItems do
			if BMTGlobal.current[realmName].items[i].marketID == foundItems[j].marketID then
				found = true
				if BMTGlobal.current[realmName].items[i].itemID ~= foundItems[j].itemID then
					itemIDMatches = false
				end
				if BMTGlobal.current[realmName].items[i].timeLeft < foundItems[j].timeLeft then
					timeLeftWentUp = true
				end
				if BMTGlobal.current[realmName].items[i].currBid > foundItems[j].currBid then
					goldWentDown = true
				end
				if BMTGlobal.current[realmName].items[i].numBids > foundItems[j].numBids then
					numBidsWentDown = true
				end
				break
			end
		end
		if not found then
			BMT_MoveEntryToHistory(realmName, i)
			currDBSize=currDBSize-1
			i=i-1
		end
		if found and (not itemIDMatches or timeLeftWentUp or goldWentDown or numBidsWentDown) then
			BMT_MoveEntryToHistory(realmName, i)
			currDBSize=currDBSize-1
			i=i-1
		end
		i=i+1
	end
end

function BMT_MoveEntryToHistory(realmName, index)
	BMT:Debug("TODO: Moving  "..BMTGlobal.current[realmName].items[index].itemID.." for realm "..realmName.." to history DB")
	table.remove(BMTGlobal.current[realmName].items, index)
end

-- called only when receiving a force update sync or searching for items outself (after which we might get a force update sync that will narrow predicted times even further)
function BMT_AddOrUpdateExisting(realmName, foundItems, fromSync)
	local currDate = date("%d/%m/%y");
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	for i=1, #foundItems do
		local found = false
		for j=1, #BMTGlobal.current[realmName].items do
			-- no need to check if names etc. are the same, as it has been done before
			-- therefore, if we've found two items with the same marketID, we are sure that this is still the same item
			-- if something breaks here, check BMT_MoveAllOldEntriesToHistory criterias
			if foundItems[i].marketID == BMTGlobal.current[realmName].items[j].marketID then
				found = true
				-- if receiving a sync, we trust it (skip the ifs)
				-- if searching ourself, we need to set the highest bidder again if gold didn't change (otherwise it will reset to "")
				-- if we are the highest bidder, none of those cases will be true, therefore foundItems[i].highestBidder will
				-- already have our name
				if not fromSync then
					if foundItems[i].currBid == BMTGlobal.current[realmName].items[j].currBid then
						foundItems[i].highestBidder = BMTGlobal.current[realmName].items[j].highestBidder
					end
				end
				-- other fields that we don't want overriden by new/default should go here
				-- if timeLeft didn't change, then restore old value, otherwise new values will be written into current db
				if foundItems[i].timeLeft >= BMTGlobal.current[realmName].items[j].timeLeft then
					foundItems[i].lastTimeLeftChangeDate = BMTGlobal.current[realmName].items[j].lastTimeLeftChangeDate
					foundItems[i].lastTimeLeftChangeTime = BMTGlobal.current[realmName].items[j].lastTimeLeftChangeTime
				end
				-- recalculate predicted times here
				if fromSync then
					BMT_RecalculateAndSetPredictedTimes(realmName, j, foundItems[i].timeLeft, currDate,  currTime, true)
				else
					BMT_RecalculateAndSetPredictedTimes(realmName, j, foundItems[i].timeLeft, currDate,  currTime, false)
				end
				-- if our old predictions are better, leave them - otherwise set default
				if foundItems[i].predictedLow <= foundItems[i].predictedHigh then
					if BMTGlobal.current[realmName].items[j].predictedLow > foundItems[i].predictedLow then
						foundItems[i].predictedLow = BMTGlobal.current[realmName].items[j].predictedLow
					end
					if BMTGlobal.current[realmName].items[j].predictedHigh < foundItems[i].predictedHigh then
						foundItems[i].predictedHigh = BMTGlobal.current[realmName].items[j].predictedHigh
					end
				end
				-- finally set all the fields in the DB
				BMTGlobal.current[realmName].items[j] = foundItems[i]
			end
		end
		-- adding new
		if found == false then
			BMT:Debug("Adding new entry: "..foundItems[i].itemID)
			BMTGlobal.current[realmName].items[#BMTGlobal.current[realmName].items+1] = foundItems[i]
		end
	end
end

-- called only when searching BMAH by yourself
function BMT_NewSearchHandler()
	-- check if Blizzard's black market ui is still opened (that's the problem with Blizzard's BMAH UI)
	if(not BlackMarketFrame:IsShown()) then return end
	local realmName = GetRealmName()
	local playerName = UnitName("player")
	local _, playerClass = UnitClass("player")
	local playerGuild = GetGuildInfo("player")
	local currHour, currMinute = GetGameTime()
	local currTime = 60*currHour+currMinute
	local currDate = date("%d/%m/%y")
	local numItems = C_BlackMarket.GetNumItems()
	-- TODO: Check what happens if numitems==0
	
	local updatedMinAgo = 0;
	updatedMinAgo = BMT_CalculateDateAndTimeDiff(BMTGlobal.current[realmName].updateInfo.updateDate, BMTGlobal.current[realmName].updateInfo.updateTime, date("%d/%m/%y"), currTime)
	local foundItems = {}
	for i=1, numItems do
		local name, texture, quantity, itemType, _, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = C_BlackMarket.GetItemInfoByIndex(i)
		local splitLink = {strsplit(":", link)}
		local itemID = tonumber(splitLink[2])
		foundItems[i] = {}
		foundItems[i].itemID = itemID
		foundItems[i].timeLeft = timeLeft;
		foundItems[i].predictedLow, foundItems[i].predictedHigh = BMT_PredictedTimeGetDefaults(timeLeft)
		-- minutes after last search time that the predictions have been updated after
		foundItems[i].predictionsUpdateDiff = 0
		foundItems[i].marketID = marketID;
		-- date and time of last blizzard's timeLeft change (used to find outdated auctions)
		foundItems[i].lastTimeLeftChangeDate = currDate
		foundItems[i].lastTimeLeftChangeTime = currTime
		foundItems[i].minBid = minBid
		foundItems[i].currBid = currBid
		foundItems[i].minIncrement = minIncrement
		foundItems[i].numBids = numBids
		foundItems[i].quantity = quantity
		local highestBidder = ""
		if youHaveHighBid then
			highestBidder = UnitName("player")
		end
		foundItems[i].highestBidder = highestBidder
	end
	
	-- if we have same amount of items and last update was done less than 1m ago then throttle
	-- there could be more conditions checked, like if any timeLeft/numBids/currBid changed, but for now this will do
	-- NOTE: numItems here is blizzard's numItems, our is returned by BMT_GetNumItems()
	if updatedMinAgo < 1 and not BMT_DidItemsChange(foundItems, numItems) then return end
	
	-- move old entries to history
	BMT_MoveAllOldEntriesToHistory(realmName, foundItems, currDate, currTime)
	-- at this point we should have no old items in current db, update existing or add new ones
	BMT_AddOrUpdateExisting(realmName, foundItems, false)
	
	BMT_SetNewUpdater(realmName, playerName, playerClass, playerGuild, currTime, currDate);
	-- update the list in case the addon's browser is opened
	-- bottom text is already updated in BMT_SetNewUpdater()
	if(BMTFrame:IsShown()) then
		BMTScrollFrame_Update();
	end
	-- sync to others
	-- UNCOMMENT
		--BMTSyncing_BroadcastCurrentSearch()
end

-- TODO: Implement
function BMT_DidItemsChange(foundItems, foundNumItems)
	return true
end

function BMT_GetNumItems(realmName)
	return #BMTGlobal.current[realmName].items
end

function BMT_PredictedTimeGetDefaults(timeLeft)
	-- completed
	if(timeLeft == TIME_LEFT_COMPLETED) then
		return 0,0;
	else
		-- example: timeLeft=1 is 0-30m, BMT_TIME_LEFT[1]=0, BMT_TIME_LEFT[2]=30
		return BMT_TIME_LEFT[timeLeft], BMT_TIME_LEFT[timeLeft+1]+1;
	end
end

-- returns difference between given poits in time in minutes (absolute)
-- date is a string in "dd/mm/yy" format, time is a number in 60*hour+minutes format
function BMT_CalculateDateAndTimeDiff(firstDate, firstTime, secondDate, secondTime)
	local t = {};
	local day, month, year = strsplit("/", firstDate);
	t["year"] = tonumber(year) + 2000;
	t["month"] = tonumber(month);
	t["day"] = tonumber(day);
	t["hour"] = firstTime / 60;
	t["min"] = firstTime % 60;
	t["sec"] = 0;
	local first = time(t);
	day, month, year = strsplit("/", secondDate);
	t["year"] = tonumber(year) + 2000;
	t["month"] = tonumber(month);
	t["day"] = tonumber(day);
	t["hour"] = secondTime / 60;
	t["min"] = secondTime % 60;
	t["sec"] = 0;
	local second = time(t);
	return (math.floor(math.abs(first-second)/60));
end

-- returns "smaller" if first date/time is smaller, "equal" if they are equal, "greater" if first is greater.
-- date is a string in "dd/mm/yy" format, time is a number in 60*hour+minutes format
function BMT_CompareDateAndTime(firstDate, firstTime, secondDate, secondTime)
	local firstDay, firstMonth, firstYear = strsplit("/", firstDate);
	local secondDay, secondMonth, secondYear = strsplit("/", secondDate);
	if(firstYear > secondYear) then return "greater"
	elseif(firstYear < secondYear) then return "smaller"
	else
		if(firstMonth > secondMonth) then return "greater"
		elseif(firstMonth < secondMonth) then return "smaller"
		else
			if(firstDay > secondDay) then return "greater"
			elseif(firstDay < secondDay) then return "smaller"
			else
				if(firstTime > secondTime) then return "greater"
				elseif(firstTime < secondTime) then return "smaller"
				else return "equal"
				end
			end
		end
	end
end
